@@ -154,6 +154,40 @@ start_server {tags {"modules"}} {
         config_set list-max-listpack-size $original_config
     }
 
+    test {Module list - KEYSIZES is updated as expected} {
+        proc run_cmd_verify_hist {cmd expOutput {retries 1}} {
+            proc K {} {return [string map { "db0_distrib_lists_items" "db0_LIST" "# Keysizes" "" " " "" "\n" "" "\r" "" } [r info keysizes] ]}
+            uplevel 1 $cmd    
+            wait_for_condition 50 $retries {
+                $expOutput eq [K]
+            } else { fail "Expected: \n`$expOutput`\n Actual:\n`[K]`.\nFailed after command: $cmd" }
+        }
+
+        r select 0
+
+        # RedisModule_ListPush & RedisModule_ListDelete
+        run_cmd_verify_hist {r flushall} {}
+        run_cmd_verify_hist {r list.insert L1 0 foo} {db0_LIST:1=1}
+        run_cmd_verify_hist {r list.insert L1 0 bla} {db0_LIST:2=1}
+        run_cmd_verify_hist {r list.delete L1 0} {db0_LIST:1=1}
+        run_cmd_verify_hist {r list.delete L1 0} {}
+        
+
+        # RedisModule_ListSet & RedisModule_ListDelete
+        run_cmd_verify_hist {r list.insert L1 0 foo} {db0_LIST:1=1}
+        run_cmd_verify_hist {r list.set L1 0 bar} {db0_LIST:1=1}
+        run_cmd_verify_hist {r list.set L1 0 baz} {db0_LIST:1=1}
+        run_cmd_verify_hist {r list.delete L1 0} {}
+
+        # Check lazy expire
+        r debug set-active-expire 0
+        run_cmd_verify_hist {r list.insert L1 0 foo} {db0_LIST:1=1}
+        run_cmd_verify_hist {r pexpire L1 1} {db0_LIST:1=1}
+        run_cmd_verify_hist {after 5} {db0_LIST:1=1}        
+        r debug set-active-expire 1
+        run_cmd_verify_hist {after 5} {} 50
+    }
+    
     test "Unload the module - list" {
         assert_equal {OK} [r module unload list]
     }
